from __future__ import annotations

from typing import TYPE_CHECKING, Callable, Iterable

import numpy as np

from zodipy.component import ComponentLabel
from zodipy.component_params import RRM

if TYPE_CHECKING:
    import numpy.typing as npt


R_0 = np.finfo(np.float64).eps
R_KUIPER_BELT = 30
R_EARTH = 1
R_JUPITER = 5.2

DIRBE_CUTOFFS: dict[ComponentLabel, tuple[float | np.float64, float | np.float64]] = {
    ComponentLabel.CLOUD: (R_0, R_JUPITER),
    ComponentLabel.BAND1: (R_0, R_JUPITER),
    ComponentLabel.BAND2: (R_0, R_JUPITER),
    ComponentLabel.BAND3: (R_0, R_JUPITER),
    ComponentLabel.RING: (R_EARTH - 0.2, R_EARTH + 0.2),
    ComponentLabel.FEATURE: (R_EARTH - 0.3, R_EARTH + 0.3),
}

RRM_CUTOFFS: dict[ComponentLabel, tuple[float | np.float64, float | np.float64]] = {
    ComponentLabel.FAN: (R_0, RRM[ComponentLabel.FAN].R_outer),  # type: ignore
    ComponentLabel.INNER_NARROW_BAND: (
        RRM[ComponentLabel.INNER_NARROW_BAND].R_inner,  # type: ignore
        RRM[ComponentLabel.INNER_NARROW_BAND].R_outer,  # type: ignore
    ),
    ComponentLabel.OUTER_NARROW_BAND: (
        RRM[ComponentLabel.OUTER_NARROW_BAND].R_inner,  # type: ignore
        RRM[ComponentLabel.OUTER_NARROW_BAND].R_outer,  # type: ignore
    ),
    ComponentLabel.BROAD_BAND: (
        RRM[ComponentLabel.BROAD_BAND].R_inner,  # type: ignore
        RRM[ComponentLabel.BROAD_BAND].R_outer,  # type: ignore
    ),
    ComponentLabel.COMET: (
        RRM[ComponentLabel.COMET].R_inner,  # type: ignore
        RRM[ComponentLabel.COMET].R_outer,  # type: ignore
    ),
    ComponentLabel.INTERSTELLAR: (R_0, R_KUIPER_BELT),
    ComponentLabel.RING_RRM: DIRBE_CUTOFFS[ComponentLabel.RING],
    ComponentLabel.FEATURE_RRM: DIRBE_CUTOFFS[ComponentLabel.FEATURE],
}


COMPONENT_CUTOFFS = {**DIRBE_CUTOFFS, **RRM_CUTOFFS}


def integrate_leggauss(
    func: Callable[[float], npt.NDArray[np.float64]],
    points: npt.NDArray[np.float64],
    weights: npt.NDArray[np.float64],
) -> npt.NDArray[np.float64]:
    """Integrate a function using Gauss-Laguerre quadrature."""
    return np.squeeze(sum(func(x) * w for x, w in zip(points, weights)))


def get_sphere_intersection(
    obs_pos: npt.NDArray[np.float64],
    unit_vectors: npt.NDArray[np.float64],
    cutoff: float | np.float64,
) -> npt.NDArray[np.float64]:
    """Returns the distance from the observer to a heliocentric sphere with radius `cutoff`."""
    x_0, y_0, z_0 = obs_pos
    r_obs = np.sqrt(x_0**2 + y_0**2 + z_0**2)
    if (r_obs > cutoff).any():
        return np.full(obs_pos.shape[-1], np.finfo(float).eps)

    u_x, u_y, u_z = unit_vectors
    lon = np.arctan2(u_y, u_x)
    lat = np.arcsin(u_z)

    cos_lat = np.cos(lat)
    b = 2 * (x_0 * cos_lat * np.cos(lon) + y_0 * cos_lat * np.sin(lon))
    c = r_obs**2 - cutoff**2

    q = -0.5 * b * (1 + np.sqrt(b**2 - 4 * c) / np.abs(b))

    return np.maximum(q, c / q)


def get_line_of_sight_range(
    components: Iterable[ComponentLabel],
    unit_vectors: npt.NDArray[np.float64],
    obs_pos: npt.NDArray[np.float64],
) -> tuple[
    dict[ComponentLabel, npt.NDArray[np.float64]],
    dict[ComponentLabel, npt.NDArray[np.float64]],
]:
    """Return the component-wise start- and end-point of each line of sight."""
    start = {
        comp: get_sphere_intersection(obs_pos, unit_vectors, COMPONENT_CUTOFFS[comp][0])
        for comp in components
    }
    stop = {
        comp: get_sphere_intersection(obs_pos, unit_vectors, COMPONENT_CUTOFFS[comp][1])
        for comp in components
    }
    return start, stop



from __future__ import annotations

import abc
from dataclasses import dataclass, field
from enum import Enum
from typing import TYPE_CHECKING

import numpy as np

if TYPE_CHECKING:
    import numpy.typing as npt


@dataclass
class ZodiacalComponent(abc.ABC):
    """Base class for storing common model parameters for zodiacal components.

    Args:
        x_0: x-offset from the Sun in heliocentric ecliptic coordinates [AU].
        y_0: y-offset from the Sun in heliocentric ecliptic coordinates [AU].
        z_0: z-offset from the Sun in heliocentric ecliptic coordinates [AU].
        i: Inclination with respect to the ecliptic plane [deg].
        Omega: Ascending node [deg].

    """

    x_0: float
    y_0: float
    z_0: float
    i: float
    Omega: float

    X_0: npt.NDArray[np.float64] = field(init=False)
    sin_i_rad: float = field(init=False)
    cos_i_rad: float = field(init=False)
    sin_Omega_rad: float = field(init=False)
    cos_Omega_rad: float = field(init=False)

    def __post_init__(self) -> None:
        self.X_0 = np.expand_dims([self.x_0, self.y_0, self.z_0], axis=-1)
        self.sin_i_rad = np.sin(np.radians(self.i))
        self.cos_i_rad = np.cos(np.radians(self.i))
        self.sin_Omega_rad = np.sin(np.radians(self.Omega))
        self.cos_Omega_rad = np.cos(np.radians(self.Omega))


@dataclass
class Cloud(ZodiacalComponent):
    """DIRBE diffuse cloud.

    Args:
        n_0: Density at 1 AU.
        alpha: Radial power-law exponent.
        beta: Vertical shape parameter.
        gamma: Vertical power-law exponent.
        mu: Widening parameter for the modified fan.

    """

    n_0: float
    alpha: float
    beta: float
    gamma: float
    mu: float


@dataclass
class Band(ZodiacalComponent):
    """DIRBE asteroidal dust band.

    Args:
        n_0: Density at 3 AU.
        delta_zeta: Shape parameter [deg].
        v: Shape parameter.
        p: Shape parameter.
        delta_r: Inner radial cutoff.

    """

    n_0: float
    delta_zeta: float
    v: float
    p: float
    delta_r: float
    delta_zeta_rad: float = field(init=False)

    def __post_init__(self) -> None:
        super().__post_init__()
        self.delta_zeta_rad = np.radians(self.delta_zeta)


@dataclass
class Ring(ZodiacalComponent):
    """DIRBE circum-solar ring (excluding the Earth-trailing Feature).

    Args:
        n_0: Density at 1 AU.
        R: Radius of the peak density.
        sigma_r: Radial dispersion.
        sigma_z: Vertical dispersion.

    """

    n_0: float
    R: float
    sigma_r: float
    sigma_z: float


@dataclass
class Feature(ZodiacalComponent):
    """DIRBE Earth-trailing Feature.

    Args:
        n_0: Density at 1 AU.
        R: Radius of the peak density.
        sigma_r: Radial dispersion.
        sigma_z: Vertical dispersion.
        theta: Longitude with respect to Earth [deg].
        sigma_theta: Longitude dispersion [deg].

    """

    n_0: float
    R: float
    sigma_r: float
    sigma_z: float
    theta: float
    sigma_theta: float
    theta_rad: float = field(init=False)
    sigma_theta_rad: float = field(init=False)

    def __post_init__(self) -> None:
        super().__post_init__()
        self.theta_rad = np.radians(self.theta)
        self.sigma_theta_rad = np.radians(self.sigma_theta)


@dataclass
class Fan(ZodiacalComponent):
    """RRM fan."""

    gamma: float
    Z_0: float
    Q: float
    P: float
    R_outer: float


@dataclass
class Comet(ZodiacalComponent):
    """RRM comet."""

    gamma: float
    Z_0: float
    P: float
    amp: float
    R_inner: float
    R_outer: float


@dataclass
class Interstellar(ZodiacalComponent):
    """RRM interstellar dust."""

    amp: float


@dataclass
class NarrowBand(ZodiacalComponent):
    """RRM narrow band."""

    gamma: float
    A: float
    G: float
    R_inner: float
    R_outer: float
    beta_nb: float


@dataclass
class BroadBand(ZodiacalComponent):
    """RRM broad band."""

    gamma: float
    A: float
    R_inner: float
    R_outer: float
    beta_bb: float
    sigma_bb: float


@dataclass
class RingRRM(Ring):
    """RRM circum-solar ring."""

    A: float


@dataclass
class FeatureRRM(Feature):
    """RRM Earth-trailing Feature."""

    A: float


class ComponentLabel(Enum):
    """Labels representing the components in the DIRBE model."""

    CLOUD = "cloud"
    BAND1 = "band1"
    BAND2 = "band2"
    BAND3 = "band3"
    RING = "ring"
    FEATURE = "feature"
    FAN = "fan"
    COMET = "comet"
    INTERSTELLAR = "interstellar"
    INNER_NARROW_BAND = "inner_narrow_band"
    OUTER_NARROW_BAND = "outer_narrow_band"
    BROAD_BAND = "broad_band"
    RING_RRM = "ring_rrm"
    FEATURE_RRM = "feature_rrm"



from zodipy.model import Model
from zodipy.model_registry import model_registry
from zodipy.number_density import grid_number_density

__all__ = (
    "Model",
    "grid_number_density",
    "model_registry",
)



from __future__ import annotations

from copy import deepcopy
from typing import Any, Callable, TypeVar, Union

import numpy as np
import numpy.typing as npt
from astropy import units
from scipy import integrate, interpolate

from zodipy.component import ComponentLabel
from zodipy.zodiacal_light_model import RRM, Kelsall, ZodiacalLightModel

CompParamDict = dict[ComponentLabel, dict[str, Any]]
CommonParamDict = dict[str, Any]
UnpackedModelDicts = tuple[CompParamDict, CommonParamDict]
T = TypeVar("T", bound=ZodiacalLightModel)
UnpackModelCallable = Callable[[units.Quantity, Union[units.Quantity, None], T], UnpackedModelDicts]


def interp_and_unpack_kelsall(
    wavelengths: units.Quantity,
    weights: units.Quantity | None,
    model: Kelsall,
) -> UnpackedModelDicts:
    """InterplantaryDustModelToDicts implementation for Kelsall model."""
    model_spectrum = deepcopy(model.spectrum)
    wavelengths = wavelengths.to(model.spectrum.unit, equivalencies=units.spectral())

    comp_params: dict[ComponentLabel, dict[str, Any]] = {}
    common_params: dict[str, Any] = {
        "T_0": model.T_0,
        "delta": model.delta,
    }

    for comp_label in model.comps:
        comp_params[comp_label] = {}
        comp_params[comp_label]["emissivity"] = interp_spectral_param(
            wavelengths,
            weights,
            model_spectrum,
            spectral_parameter=model.emissivities[comp_label],
        )
        if model.albedos is not None:
            comp_params[comp_label]["albedo"] = interp_spectral_param(
                wavelengths,
                weights,
                model_spectrum,
                spectral_parameter=model.albedos[comp_label],
            )
        else:
            comp_params[comp_label]["albedo"] = 0

    common_params["C1"] = (
        interp_spectral_param(
            wavelengths,
            weights,
            model_spectrum,
            spectral_parameter=model.C1,
            use_nearest=True,
        )
        if model.C1 is not None
        else 0
    )
    common_params["C2"] = (
        interp_spectral_param(
            wavelengths,
            weights,
            model_spectrum,
            spectral_parameter=model.C2,
            use_nearest=True,
        )
        if model.C2 is not None
        else 0
    )

    common_params["C3"] = (
        interp_spectral_param(
            wavelengths,
            weights,
            model_spectrum,
            spectral_parameter=model.C3,
            use_nearest=True,
        )
        if model.C3 is not None
        else 0
    )
    if model.solar_irradiance is None:
        common_params["solar_irradiance"] = 0
    else:
        common_params["solar_irradiance"] = interp_spectral_param(
            wavelengths,
            weights,
            model_spectrum,
            spectral_parameter=model.solar_irradiance,
        )

    return comp_params, common_params


def interp_and_unpack_rrm(
    wavelengths: units.Quantity,
    weights: units.Quantity | None,
    model: RRM,
) -> UnpackedModelDicts:
    """InterplantaryDustModelToDicts implementation for Kelsall model."""
    model_spectrum = deepcopy(model.spectrum)
    wavelengths = wavelengths.to(model.spectrum.unit, equivalencies=units.spectral())

    comp_params: dict[ComponentLabel, dict[str, Any]] = {}
    common_params: dict[str, Any] = {}

    for comp_label in model.comps:
        comp_params[comp_label] = {}
        comp_params[comp_label]["T_0"] = model.T_0[comp_label]
        comp_params[comp_label]["delta"] = model.delta[comp_label]

    calibration = interp_spectral_param(
        wavelengths, weights, model_spectrum, spectral_parameter=model.calibration
    )
    calibration_quantity = units.Quantity(calibration, unit=units.MJy / units.AU)
    common_params["calibration"] = calibration_quantity.to_value(units.Jy / units.cm)
    return comp_params, common_params


def interp_spectral_param(
    wavelengths: units.Quantity,
    weights: units.Quantity | None,
    model_spectrum: units.Quantity,
    spectral_parameter: npt.ArrayLike,
    use_nearest: bool = False,
) -> npt.NDArray:
    """Interpolate a spectral parameters."""
    paramameter = np.asarray(spectral_parameter)

    if not np.array_equal(model_spectrum.value, np.sort(model_spectrum.value)):
        model_spectrum = np.flip(model_spectrum)
        paramameter = np.flip(paramameter)

    if use_nearest:
        interped_param = interpolate.interp1d(model_spectrum.value, paramameter, kind="nearest")(
            wavelengths.value
        )
    else:
        interped_param = np.interp(wavelengths.value, model_spectrum.value, paramameter)

    if weights is not None:
        return integrate.trapezoid(weights.value * interped_param, x=wavelengths.value)
    return interped_param


interp_and_unpack_func_mapping: dict[type[ZodiacalLightModel], UnpackModelCallable] = {
    Kelsall: interp_and_unpack_kelsall,
    RRM: interp_and_unpack_rrm,
}


def get_model_interp_func(
    model: ZodiacalLightModel,
) -> UnpackModelCallable:
    """Get the appropriate parameter unpacker for the model."""
    return interp_and_unpack_func_mapping[type(model)]



from __future__ import annotations

import copy
import inspect
from dataclasses import asdict
from functools import partial
from typing import TYPE_CHECKING, Callable, Mapping, Protocol

import numpy as np
import numpy.typing as npt

from zodipy.bodies import get_earthpos_inst
from zodipy.component import (
    Band,
    BroadBand,
    Cloud,
    Comet,
    ComponentLabel,
    Fan,
    Feature,
    FeatureRRM,
    Interstellar,
    NarrowBand,
    Ring,
    RingRRM,
    ZodiacalComponent,
)
from zodipy.model_registry import model_registry
from zodipy.zodiacal_light_model import ZodiacalLightModel

if TYPE_CHECKING:
    from astropy import time, units


"""The density functions for the different types of components.

Common for all of these is that the first argument will be `X_helio` (the line
of sight from the observer towards a point on the sky) and the remaining arguments
will be parameters that are set by the `Component` subclasses. These arguments are
unpacked automatically, so for a `ComputeDensityFunc` to work, the mapped `Component`
class must have all the parameters as instance variables.

"""
ComputeDensityFunc = Callable[..., npt.NDArray[np.float64]]


def cloud_number_density(
    X_helio: npt.NDArray[np.float64],
    X_0: npt.NDArray[np.float64],
    sin_Omega_rad: float,
    cos_Omega_rad: float,
    sin_i_rad: float,
    cos_i_rad: float,
    n_0: float,
    mu: float,
    alpha: float,
    beta: float,
    gamma: float,
) -> npt.NDArray[np.float64]:
    """Density of the diffuse cloud (see Eq (6). in K98)."""
    X_cloud = X_helio - X_0
    R_cloud = np.sqrt(X_cloud[0] ** 2 + X_cloud[1] ** 2 + X_cloud[2] ** 2)
    Z_cloud = (
        X_cloud[0] * sin_Omega_rad * sin_i_rad
        - X_cloud[1] * cos_Omega_rad * sin_i_rad
        + X_cloud[2] * cos_i_rad
    )

    zeta = np.abs(Z_cloud / R_cloud)

    g = np.where(zeta < mu, zeta**2 / (2 * mu), zeta - (mu / 2))

    return n_0 * R_cloud**-alpha * np.exp(-beta * g**gamma)


def band_number_density(
    X_helio: npt.NDArray[np.float64],
    X_0: npt.NDArray[np.float64],
    sin_Omega_rad: float,
    cos_Omega_rad: float,
    sin_i_rad: float,
    cos_i_rad: float,
    n_0: float,
    delta_zeta_rad: float,
    p: float,
    v: float,
    delta_r: float,
) -> npt.NDArray[np.float64]:
    """Density of the dust bands (see Eq. (8) in K98)."""
    X_band = X_helio - X_0
    R_band = np.sqrt(X_band[0] ** 2 + X_band[1] ** 2 + X_band[2] ** 2)

    Z_band = (
        X_band[0] * sin_Omega_rad * sin_i_rad
        - X_band[1] * cos_Omega_rad * sin_i_rad
        + X_band[2] * cos_i_rad
    )

    zeta = np.abs(Z_band / R_band)
    zeta_over_delta_zeta = zeta / delta_zeta_rad
    term1 = 3 * n_0 / R_band
    term2 = np.exp(-(zeta_over_delta_zeta**6))

    # Differs from eq 8 in K98 by a factor of 1/self.v. See Planck XIV
    # section 4.1.2.
    term3 = 1 + (zeta_over_delta_zeta**p) / v

    term4 = 1 - np.exp(-((R_band / delta_r) ** 20))

    return term1 * term2 * term3 * term4


def ring_number_density(
    X_helio: npt.NDArray[np.float64],
    X_0: npt.NDArray[np.float64],
    sin_Omega_rad: float,
    cos_Omega_rad: float,
    sin_i_rad: float,
    cos_i_rad: float,
    n_0: float,
    R: float,
    sigma_r: float,
    sigma_z: float,
) -> npt.NDArray[np.float64]:
    """Density of the circum-solar ring (see Eq. (9) in K98)."""
    X_ring = X_helio - X_0
    R_ring = np.sqrt(X_ring[0] ** 2 + X_ring[1] ** 2 + X_ring[2] ** 2)

    Z_ring = (
        X_ring[0] * sin_Omega_rad * sin_i_rad
        - X_ring[1] * cos_Omega_rad * sin_i_rad
        + X_ring[2] * cos_i_rad
    )
    # Differs from eq 9 in K98 by a factor of 1/2 in the first and last
    # term. See Planck 2013 XIV, section 4.1.3.
    term1 = -((R_ring - R) ** 2) / sigma_r**2
    term2 = np.abs(Z_ring) / sigma_z

    return n_0 * np.exp(term1 - term2)


def feature_number_density(
    X_helio: npt.NDArray[np.float64],
    X_0: npt.NDArray[np.float64],
    X_earth: npt.NDArray[np.float64],
    sin_Omega_rad: float,
    cos_Omega_rad: float,
    sin_i_rad: float,
    cos_i_rad: float,
    n_0: float,
    R: float,
    sigma_r: float,
    sigma_z: float,
    theta_rad: float,
    sigma_theta_rad: float,
) -> npt.NDArray[np.float64]:
    """Density of the Earth-trailing feature (see Eq. (9) in K98)."""
    X_feature = X_helio - X_0
    R_feature = np.sqrt(X_feature[0] ** 2 + X_feature[1] ** 2 + X_feature[2] ** 2)

    Z_feature = (
        X_feature[0] * sin_Omega_rad * sin_i_rad
        - X_feature[1] * cos_Omega_rad * sin_i_rad
        + X_feature[2] * cos_i_rad
    )

    X_earth_comp = X_earth - X_0
    theta_comp = np.arctan2(X_feature[1], X_feature[0]) - np.arctan2(
        X_earth_comp[1], X_earth_comp[0]
    )

    delta_theta = theta_comp - theta_rad
    delta_theta = (delta_theta + np.pi) % (2 * np.pi) - np.pi

    # Differs from eq 9 in K98 by a factor of 1/2 in the first and last
    # term. See Planck 2013 XIV, section 4.1.3.
    exp_term = (R_feature - R) ** 2 / sigma_r**2
    exp_term += np.abs(Z_feature) / sigma_z
    exp_term += delta_theta**2 / sigma_theta_rad**2

    return n_0 * np.exp(-exp_term)


def fan_number_density(
    X_helio: npt.NDArray[np.float64],
    X_0: npt.NDArray[np.float64],
    sin_Omega_rad: float,
    cos_Omega_rad: float,
    sin_i_rad: float,
    cos_i_rad: float,
    Q: float,
    P: float,
    gamma: float,
    Z_0: float,
    R_outer: float,
) -> npt.NDArray[np.float64]:
    """Density of the fan (see Eq (3). in RRM)."""
    X_fan = X_helio - X_0
    R_fan = np.sqrt(X_fan[0] ** 2 + X_fan[1] ** 2 + X_fan[2] ** 2)

    density = np.zeros_like(R_fan)
    indices = R_fan <= R_outer
    X_fan_filtered = X_fan[:, indices]
    R_fan_filtered = R_fan[indices]

    Z_fan = (
        X_fan_filtered[0] * sin_Omega_rad * sin_i_rad
        - X_fan_filtered[1] * cos_Omega_rad * sin_i_rad
        + X_fan_filtered[2] * cos_i_rad
    )
    sin_beta = Z_fan / R_fan_filtered
    beta = np.arcsin(sin_beta)
    Z_fan_abs = np.abs(Z_fan)
    epsilon = np.where(Z_fan_abs < Z_0, 2 - (Z_fan_abs / Z_0), 1)
    f = np.cos(beta) ** Q * np.exp(-P * np.sin(np.abs(beta) ** epsilon))

    density[indices] = (R_fan_filtered ** (-gamma)) * f
    return density


def comet_number_density(
    X_helio: npt.NDArray[np.float64],
    X_0: npt.NDArray[np.float64],
    sin_Omega_rad: float,
    cos_Omega_rad: float,
    sin_i_rad: float,
    cos_i_rad: float,
    gamma: float,
    Z_0: float,
    P: float,
    amp: float,
    R_inner: float,
    R_outer: float,
) -> npt.NDArray[np.float64]:
    """Density of the fan (see Eq (3). in RRM)."""
    X_comet = X_helio - X_0
    R_comet = np.sqrt(X_comet[0] ** 2 + X_comet[1] ** 2 + X_comet[2] ** 2)

    density = np.zeros_like(R_comet)
    indices = np.logical_and(R_comet >= R_inner, R_comet <= R_outer)
    X_comet_filtered = X_comet[:, indices]
    R_comet_filtered = R_comet[indices]

    Z_comet = (
        X_comet_filtered[0] * sin_Omega_rad * sin_i_rad
        - X_comet_filtered[1] * cos_Omega_rad * sin_i_rad
        + X_comet_filtered[2] * cos_i_rad
    )
    sin_beta = Z_comet / R_comet_filtered
    beta = np.arcsin(sin_beta)
    Z_fan_abs = np.abs(Z_comet)
    epsilon = np.where(Z_fan_abs < Z_0, 2 - (Z_fan_abs / Z_0), 1)
    f = np.exp(-P * np.sin(np.abs(beta) ** epsilon))

    density[indices] = amp * (R_comet_filtered ** (-gamma)) * f
    return density


def interstellar_number_density(
    X_helio: npt.NDArray[np.float64],
    amp: float,
) -> npt.NDArray[np.float64]:
    """Interstellar constant number density."""
    return np.array([amp])


def narrow_band_number_density(
    X_helio: npt.NDArray[np.float64],
    X_0: npt.NDArray[np.float64],
    sin_Omega_rad: float,
    cos_Omega_rad: float,
    sin_i_rad: float,
    cos_i_rad: float,
    beta_nb: float,
    G: float,
    gamma: float,
    A: float,
    R_inner: float,
    R_outer: float,
) -> npt.NDArray[np.float64]:
    """Density of the fan (see Eq (4). in RRM)."""
    X_nb = X_helio - X_0
    R_nb = np.sqrt(X_nb[0] ** 2 + X_nb[1] ** 2 + X_nb[2] ** 2)

    density = np.zeros_like(R_nb)
    indices = np.logical_and(R_nb >= R_inner, R_nb <= R_outer)

    X_nb_filtered = X_nb[:, indices]
    R_nb_filtered = R_nb[indices]

    Z_nb = (
        X_nb_filtered[0] * sin_Omega_rad * sin_i_rad
        - X_nb_filtered[1] * cos_Omega_rad * sin_i_rad
        + X_nb_filtered[2] * cos_i_rad
    )

    sin_beta = Z_nb / R_nb_filtered
    beta_abs = np.abs(np.rad2deg(np.arcsin(sin_beta)))

    f = np.where(beta_abs < beta_nb, np.exp(G * (beta_abs - beta_nb)), 0)

    density[indices] = A * ((R_nb_filtered / R_outer) ** (-gamma)) * f

    return density


def broad_band_number_density(
    X_helio: npt.NDArray[np.float64],
    X_0: npt.NDArray[np.float64],
    sin_Omega_rad: float,
    cos_Omega_rad: float,
    sin_i_rad: float,
    cos_i_rad: float,
    beta_bb: float,
    sigma_bb: float,
    gamma: float,
    A: float,
    R_inner: float,
    R_outer: float,
) -> npt.NDArray[np.float64]:
    """Density of the fan (see Eq (5). in RRM)."""
    X_bb = X_helio - X_0
    R_bb = np.sqrt(X_bb[0] ** 2 + X_bb[1] ** 2 + X_bb[2] ** 2)

    density = np.zeros_like(R_bb)
    indices = np.logical_and(R_bb >= R_inner, R_bb <= R_outer)
    X_bb_filtered = X_bb[:, indices]
    R_bb_filtered = R_bb[indices]

    Z_bb = (
        X_bb_filtered[0] * sin_Omega_rad * sin_i_rad
        - X_bb_filtered[1] * cos_Omega_rad * sin_i_rad
        + X_bb_filtered[2] * cos_i_rad
    )
    sin_beta = Z_bb / R_bb_filtered
    beta = np.rad2deg(np.arcsin(sin_beta))

    f = np.exp(-0.5 * ((beta - beta_bb) / sigma_bb) ** 2) + np.exp(
        -0.5 * ((beta + beta_bb) / sigma_bb) ** 2
    )
    density[indices] = A * ((R_bb_filtered / R_outer) ** (-gamma)) * f
    return density


def rrm_ring_number_density(
    X_helio: npt.NDArray[np.float64],
    X_0: npt.NDArray[np.float64],
    sin_Omega_rad: float,
    cos_Omega_rad: float,
    sin_i_rad: float,
    cos_i_rad: float,
    n_0: float,
    R: float,
    sigma_r: float,
    sigma_z: float,
    A: float,
) -> npt.NDArray[np.float64]:
    """RRM ring is just K98 ring with a scaling factor."""
    return A * ring_number_density(
        X_helio=X_helio,
        X_0=X_0,
        sin_Omega_rad=sin_Omega_rad,
        cos_Omega_rad=cos_Omega_rad,
        sin_i_rad=sin_i_rad,
        cos_i_rad=cos_i_rad,
        n_0=n_0,
        R=R,
        sigma_r=sigma_r,
        sigma_z=sigma_z,
    )


def rrm_feature_number_density(
    X_helio: npt.NDArray[np.float64],
    X_0: npt.NDArray[np.float64],
    X_earth: npt.NDArray[np.float64],
    sin_Omega_rad: float,
    cos_Omega_rad: float,
    sin_i_rad: float,
    cos_i_rad: float,
    n_0: float,
    R: float,
    sigma_r: float,
    sigma_z: float,
    theta_rad: float,
    sigma_theta_rad: float,
    A: float,
) -> npt.NDArray[np.float64]:
    """RRM feature is just K98 feature with a scaling factor."""
    return A * feature_number_density(
        X_helio=X_helio,
        X_0=X_0,
        sin_Omega_rad=sin_Omega_rad,
        cos_Omega_rad=cos_Omega_rad,
        sin_i_rad=sin_i_rad,
        cos_i_rad=cos_i_rad,
        n_0=n_0,
        R=R,
        sigma_r=sigma_r,
        sigma_z=sigma_z,
        X_earth=X_earth,
        sigma_theta_rad=sigma_theta_rad,
        theta_rad=theta_rad,
    )


# Mapping of implemented zodiacal component data classes and their density functions.
DENSITY_FUNCS: dict[type[ZodiacalComponent], ComputeDensityFunc] = {
    Cloud: cloud_number_density,
    Band: band_number_density,
    Ring: ring_number_density,
    Feature: feature_number_density,
    Fan: fan_number_density,
    Comet: comet_number_density,
    Interstellar: interstellar_number_density,
    NarrowBand: narrow_band_number_density,
    BroadBand: broad_band_number_density,
    RingRRM: rrm_ring_number_density,
    FeatureRRM: rrm_feature_number_density,
}


class NumberDensityFunc(Protocol):
    """Protocol for a zodiacal components number density function."""

    def __call__(self, X_helio: npt.NDArray[np.float64]) -> npt.NDArray[np.float64]:
        """Return the number density of the component at the heliocentric position."""


def get_partial_number_density_func(
    comps: Mapping[ComponentLabel, ZodiacalComponent],
) -> dict[ComponentLabel, partial[npt.NDArray[np.float64]]]:
    """Construct density partials for components.

    Return a tuple of the density expressions above which has been prepopulated with
    model and configuration parameters, leaving only the `X_helio` argument to be supplied.
    Raises exception for incorrectly defined components or component density functions.
    """
    partial_density_funcs: dict[ComponentLabel, partial[npt.NDArray[np.float64]]] = {}
    for comp_label, comp in comps.items():
        comp_dict = asdict(comp)
        func_params = inspect.signature(DENSITY_FUNCS[type(comp)]).parameters.keys()
        residual_params = [key for key in func_params if key not in comp_dict]
        try:
            residual_params.remove("X_helio")
        except ValueError as err:  # pragma: no cover
            msg = "X_helio must be be the first argument to the density function of a component."
            raise ValueError(msg) from err

        if "X_earth" in residual_params:
            residual_params.remove("X_earth")

        if residual_params:  # pragma: no cover
            msg = (
                f"Argument(s) {residual_params} required by the density function "
                f"{DENSITY_FUNCS[type(comp)]} are not provided by instance variables in "
                f"{type(comp)} or by the `computed_parameters` dict."
            )
            raise ValueError(msg)

        # Remove excess intermediate parameters from the component dict.
        comp_params = {key: value for key, value in comp_dict.items() if key in func_params}
        partial_func = partial(DENSITY_FUNCS[type(comp)], **comp_params)
        partial_density_funcs[comp_label] = partial_func

    return partial_density_funcs


def update_partial_earth_pos(
    partials: dict[ComponentLabel, partial[npt.NDArray[np.float64]]],
    earth_pos: npt.NDArray[np.float64],
) -> dict[ComponentLabel, partial[npt.NDArray[np.float64]]]:
    """Inplace populate the `X_earth` parameter in the partial density functions."""
    updated_partials = copy.deepcopy(partials)
    for partial_func in updated_partials.values():
        remaining = inspect.signature(partial_func).parameters.keys() - partial_func.keywords.keys()
        if "X_earth" in remaining:
            partial_func.keywords["X_earth"] = earth_pos
    return updated_partials


def grid_number_density(
    x: units.Quantity[units.AU],
    y: units.Quantity[units.AU],
    z: units.Quantity[units.AU],
    obstime: time.Time,
    model: str | ZodiacalLightModel = "dirbe",
    ephemeris: str = "builtin",
) -> npt.NDArray[np.float64]:
    """Return the component-wise tabulated densities of the zodiacal components for a given grid.

    Args:
        x: x-coordinates of a cartesian mesh grid.
        y: y-coordinates of a cartesian mesh grid.
        z: z-coordinates of a cartesian mesh grid.
        obstime: Time of observation. Required to compute the Earth position.
        model: String representing a built-in model or an explicit instance of a
            `ZodiacalLightModel`. Default is 'dirbe'.
        ephemeris: Solar system ephemeris to use. Default is 'builtin'.

    Returns:
        number_density_grid: The tabulated zodiacal component densities.

    """
    if isinstance(model, str):
        zodiacal_light_model = model_registry.get_model(model)
    elif isinstance(model, ZodiacalLightModel):
        zodiacal_light_model = model
    else:
        msg = "model type must be a `str` or a `ZodiacalLightModel`."
        raise TypeError(msg)

    grid = np.asarray(np.meshgrid(x, y, z))

    earthpos_xyz = get_earthpos_inst(obstime, ephemeris=ephemeris)

    # broadcasting reshapes
    for comp in zodiacal_light_model.comps.values():
        comp.X_0 = comp.X_0.reshape(3, 1, 1, 1)

    density_partials = get_partial_number_density_func(
        comps=zodiacal_light_model.comps,
    )
    density_partials = update_partial_earth_pos(
        density_partials, earthpos_xyz[:, np.newaxis, np.newaxis, np.newaxis]
    )

    number_density_grid = np.zeros((len(zodiacal_light_model.comps), *grid.shape[1:]))
    for idx, number_density_callable in enumerate(density_partials.values()):
        number_density_grid[idx] = number_density_callable(grid)

    # Revert broadcasting reshapes
    for comp in zodiacal_light_model.comps.values():
        comp.X_0 = comp.X_0.reshape(3, 1)

    return number_density_grid



from __future__ import annotations

from typing import TYPE_CHECKING, Callable

import numpy as np
import numpy.typing as npt

from zodipy.blackbody import get_dust_grain_temperature
from zodipy.scattering import get_phase_function, get_scattering_angle

if TYPE_CHECKING:
    from zodipy.number_density import NumberDensityFunc

"""
Function that return the zodiacal emission at a step along all lines of sight given
a zodiacal model.
"""
BrightnessAtStepCallable = Callable[..., npt.NDArray[np.float64]]


def kelsall_brightness_at_step(
    r: npt.NDArray[np.float64],
    start: npt.NDArray[np.float64],
    stop: npt.NDArray[np.float64],
    X_obs: npt.NDArray[np.float64],
    u_los: npt.NDArray[np.float64],
    bp_interpolation_table: npt.NDArray[np.float64],
    number_density_func: NumberDensityFunc,
    T_0: float,
    delta: float,
    emissivity: np.float64,
    albedo: np.float64,
    C1: np.float64,
    C2: np.float64,
    C3: np.float64,
    solar_irradiance: np.float64,
) -> npt.NDArray[np.float64]:
    """Kelsall uses common line of sight grid from obs to 5.2 AU."""
    # Convert the quadrature range from [-1, 1] to the true ecliptic positions
    # and back again at the end
    R_los = 0.5 * (stop - start) * r + 0.5 * (stop + start)

    X_los = R_los * u_los
    X_helio = X_los + X_obs
    R_helio = np.sqrt(X_helio[0] ** 2 + X_helio[1] ** 2 + X_helio[2] ** 2)

    temperature = get_dust_grain_temperature(R_helio, T_0, delta)
    blackbody_emission = np.interp(temperature, *bp_interpolation_table)
    emission = (1 - albedo) * (emissivity * blackbody_emission)
    if albedo != 0:
        solar_flux = solar_irradiance / R_helio**2
        scattering_angle = get_scattering_angle(R_los, R_helio, X_los, X_helio)
        phase_function = get_phase_function(scattering_angle, C1, C2, C3)
        emission += albedo * solar_flux * phase_function

    return emission * number_density_func(X_helio) * 0.5 * (stop - start)


def rrm_brightness_at_step(
    r: npt.NDArray[np.float64],
    start: npt.NDArray[np.float64],
    stop: npt.NDArray[np.float64],
    X_obs: npt.NDArray[np.float64],
    u_los: npt.NDArray[np.float64],
    bp_interpolation_table: npt.NDArray[np.float64],
    number_density_func: NumberDensityFunc,
    T_0: float,
    delta: float,
    calibration: np.float64,
) -> npt.NDArray[np.float64]:
    """RRM is implented with component specific line-of-sight grids."""
    # Convert the quadrature range from [-1, 1] to the true ecliptic positions
    # and back again at the end
    R_los = 0.5 * (stop - start) * r + 0.5 * (stop + start)

    X_los = R_los * u_los
    X_helio = X_los + X_obs
    R_helio = np.sqrt(X_helio[0] ** 2 + X_helio[1] ** 2 + X_helio[2] ** 2)

    temperature = get_dust_grain_temperature(R_helio, T_0, delta)
    blackbody_emission = np.interp(temperature, *bp_interpolation_table)

    return blackbody_emission * number_density_func(X_helio) * calibration * 0.5 * (stop - start)



from __future__ import annotations

from typing import TYPE_CHECKING

import numpy as np

if TYPE_CHECKING:
    import numpy.typing as npt


def get_scattering_angle(
    R_los: float | npt.NDArray[np.float64],
    R_helio: npt.NDArray[np.float64],
    X_los: npt.NDArray[np.float64],
    X_helio: npt.NDArray[np.float64],
) -> npt.NDArray[np.float64]:
    """Return the scattering angle between the Sun and a point along the line of sight.

    Args:
        R_los: Distance from observer to a point along the line of sight.
        R_helio: Distance from the Sun to a point along the line of sight.
        X_los: Vector pointing from the observer to a point along the line of sight.
        X_helio: Heliocentric position of point along the line of sight.

    Returns:
        Scattering angle.

    """
    cos_theta = (X_los * X_helio).sum(axis=0) / (R_los * R_helio)
    cos_theta = np.clip(cos_theta, -1, 1)
    return np.arccos(-cos_theta)


def get_phase_function(
    Theta: npt.NDArray[np.float64], C1: np.float64, C2: np.float64, C3: np.float64
) -> npt.NDArray[np.float64]:
    """Return the phase function.

    Args:
        Theta: Scattering angle.
        C1: Phase_function coefficient.
        C2: Phase_function coefficient.
        C3: Phase_function coefficient.

    Returns:
        The Phase funciton.

    """
    phase_normalization = _get_phase_normalization(C1, C2, C3)
    return phase_normalization * (C1 + C2 * Theta + np.exp(C3 * Theta))


def _get_phase_normalization(C1: np.float64, C2: np.float64, C3: np.float64) -> np.float64:
    """Return the analyitcal integral for the phase normalization factor N."""
    int_term1 = 2 * np.pi
    int_term2 = 2 * C1
    int_term3 = np.pi * C2
    int_term4 = (np.exp(C3 * np.pi) + 1) / (C3**2 + 1)
    return 1 / (int_term1 * (int_term2 + int_term3 + int_term4))
